using System;
using System.Collections;
using OSGeo.FDO.Connections;
using OSGeo.FDO.ClientServices;
using OSGeo.FDO;
using System.IO;

namespace MyFDO
{
    /// <summary>
    /// Establish connection to a specific FDO Provider
    /// </summary>
    public class FDOConnector
    {
        #region Attributes
        internal IConnection connection;
        private string _directory;
        #endregion

        #region Properties
        /// <summary>
        /// Check if the connection is opened
        /// </summary>
        public bool IsOpened 
        {
            get { return (connection!=null && connection.ConnectionState == ConnectionState.ConnectionState_Open); }
        }
        public string CurrentDirectory
        {
            get { return _directory; }
            set { 
                _directory = value; 
                if(!string.IsNullOrEmpty(_directory)) 
                    if(!_directory.EndsWith(@"\")) _directory += @"\";
            }
        }
        #endregion

        #region CTOR/DTOR
        public FDOConnector() { }
        #endregion

        #region Methods
        /// <summary>
        /// Execute open command on preset 
        /// </summary>
        /// <returns></returns>
        private bool Open()
        {
            ConnectionState connectState = connection.Open();
            switch (connectState)
            {
                case ConnectionState.ConnectionState_Busy: break;
                case ConnectionState.ConnectionState_Closed: break;
                case ConnectionState.ConnectionState_Open: break;
                case ConnectionState.ConnectionState_Pending:
                    // pending connection, retrying
                    connectState = connection.Open();
                    break;
                default:
                    return false;  //throw new FDOException(FDOExceptionCause.InvalidConnectionState);
            }
            return true;
        }
        /// <summary>
        /// Open SHP provider
        /// </summary>
        /// <param name="defaultDirectory"></param>
        /// <param name="tempDirectory"></param>
        /// <returns></returns>
        public bool OpenSHP(string defaultDirectory, string tempDirectory)
        {
            if (string.IsNullOrEmpty(defaultDirectory) || !Directory.Exists(defaultDirectory)) return false; //throw new FDOException(FDOExceptionCause.InvalidDefaultDirectory);
            if (string.IsNullOrEmpty(tempDirectory) || !Directory.Exists(tempDirectory)) return false; //throw new FDOException(FDOExceptionCause.InvalidTempDirectory);

            IConnectionManager connectionManager = FeatureAccessManager.GetConnectionManager();
            connection = connectionManager.CreateConnection(getSHPProviderName());

            IConnectionPropertyDictionary connectionPropertyDictionary;            
            connectionPropertyDictionary = connection.ConnectionInfo.ConnectionProperties;

            connectionPropertyDictionary.SetProperty("DefaultFileLocation", defaultDirectory);
            connectionPropertyDictionary.SetProperty("TemporaryFileLocation", tempDirectory);

            this.CurrentDirectory = defaultDirectory;

            return this.Open();
        }
        /// <summary>
        /// Open SDF provider
        /// </summary>
        /// <param name="fileName"></param>
        /// <param name="readOnly"></param>
        /// <returns></returns>
        public bool OpenSDF(string fileName, bool readOnly)
        {
            if (string.IsNullOrEmpty(fileName) || !(new FileInfo(fileName).Exists)) return false; //throw new FDOException(FDOExceptionCause.InvalidDefaultDirectory);
            
            IConnectionManager connectionManager = FeatureAccessManager.GetConnectionManager();
            connection = connectionManager.CreateConnection(getSDFProviderName());

            IConnectionPropertyDictionary connectionPropertyDictionary;
            connectionPropertyDictionary = connection.ConnectionInfo.ConnectionProperties;

            connectionPropertyDictionary.SetProperty("File", fileName);
            connectionPropertyDictionary.SetProperty("ReadOnly", readOnly.ToString());

            this.CurrentDirectory = Path.GetDirectoryName(fileName);

            return this.Open();
        }
        /// <summary>
        /// Open the connection
        /// </summary>
        /// <param name="provider">Provider to load</param>
        /// <param name="dl">SHP Default Directory</param>
        /// <param name="tl">SHP Temporary Directory</param>
        /// <param name="tl">SDF File</param>
        /// <param name="tl">SDF ReadOnly</param>
        public bool Open(FDOProvider provider, string dl, string tl, string f, bool ro)
        {
            switch (provider)
            {
                case FDOProvider.Shape :
                    return this.OpenSHP(dl, tl);
                case FDOProvider.SDF:
                    return this.OpenSDF(f, ro);
            }
            return false;

        }
        /// <summary>
        /// Close the connection
        /// </summary>
        public void Close()
        {
            try
            {
                if (IsOpened) connection.Close();
            }
            catch { }
        }
        
        #region Provider Methods
        /// <summary>
        /// Browse provider list to retrieve SHP Provider
        /// </summary>
        /// <returns>provider connection string</returns>
        private string getSHPProviderName()
        {
            IProviderRegistry ipr = FeatureAccessManager.GetProviderRegistry();
            ProviderCollection pc = ipr.GetProviders();
            foreach (Provider p in pc)
            {
                if (p.LibraryPath.ToLower() == "shpprovider.dll") return p.Name;
            }
            return null;
        }
        /// <summary>
        /// Browse provider list to retrieve SHP Provider
        /// </summary>
        /// <returns>provider connection string</returns>
        private string getSDFProviderName()
        {
            IProviderRegistry ipr = FeatureAccessManager.GetProviderRegistry();
            ProviderCollection pc = ipr.GetProviders();
            foreach (Provider p in pc)
            {
                if (p.LibraryPath.ToLower() == "sdfprovider.dll") return p.Name;
            }
            return null;
        }
        /// <summary>
        /// Get the providers list
        /// </summary>
        /// <returns></returns>
        public static ProviderCollection GetProviderCollection()
        {
            IProviderRegistry ipr = FeatureAccessManager.GetProviderRegistry();
            return  ipr.GetProviders();            
        }
        /// <summary>
        /// Get the providers names
        /// </summary>
        /// <returns>string array</returns>
        public static ArrayList GetProviders()
        {
            ArrayList list = new ArrayList();
            IProviderRegistry ipr = FeatureAccessManager.GetProviderRegistry();
            ProviderCollection pc = ipr.GetProviders();
            foreach (Provider p in pc)
            {
                list.Add(p.Name);
            }
            list.Sort();
            return list;
        }
        #endregion

        #endregion

    }

    #region Enums
    /// <summary>
    /// Dimensionality list
    /// </summary>
    internal enum FDODimensionality
    {
        Dimensionality_XY = 0,
        Dimensionality_Z = 1,
        Dimensionality_M = 2
    };

    /// <summary>
    /// Managed Provider list
    /// </summary>
    public enum FDOProvider
    {
        Shape, SDF
    };

    #endregion
    
    #region Exceptions
    /// <summary>
    /// FDO Exception
    /// </summary>
    public class FDOException : ApplicationException
    {
        public FDOExceptionCause Cause;
        /// <summary>
        /// FDO Exception
        /// </summary>
        /// <param name="c">FDOExceptionCause</param>
        public FDOException(FDOExceptionCause c)
        {
            Cause = c;
        }
    }
    /// <summary>
    /// Known FDO Exception causes
    /// </summary>
    public enum FDOExceptionCause {
        InvalidProvider,
        InvalidDefaultDirectory,
        InvalidTempDirectory,
        InvalidConnectionState
    }
    #endregion

}
