using System;
using System.IO;
using System.Data;
using OSGeo.FDO.Schema;
using OSGeo.FDO.Commands.Schema;
using OSGeo.FDO.Commands;
using OSGeo.FDO.Expression;
using OSGeo.FDO.Geometry;
using OSGeo.FDO.Commands.Feature;
using System.Collections.Generic;
using System.Collections;
using OSGeo.FDO.Common;

namespace MyFDO
{

    // TODO : manage SDF Files on FDO write methods (open is ok)
    // TODO : generic write methods with safe code

    /// <summary>
    /// Manage FDO files write access
    /// </summary>
    public class FDOManager
    {  
        // backup related data
        const string BACKUP_DIR = @"Backup\";
        const double BACKUP_DIR_LENGTH = 0; // keep files for X days
        const int BACKUP_KEEP = 5; // keep last X directories regardless of the date
        public bool DoBackup = true;
        private DirectoryInfo backupDir;
        private Hashtable backupedFiles;

        private bool _isOpen = false;

        private FDOConnector fdo;

        // backup for ReOpen method
        private FDOProvider source;
        private string shp_dl, shp_tl;
        private string sdf_file; bool sdf_read;


        #region FDOManager
        private FDOManager() { }
        /// <summary>
        /// Init as SHP Manager
        /// </summary>
        /// <param name="provider"></param>
        /// <param name="dl"></param>
        /// <param name="tl"></param>
        public FDOManager(string dl, string tl)
        {
            // backup connection string
            this.shp_dl = dl;
            this.shp_tl = tl;
            // open connection
            fdo = new FDOConnector();
            this._isOpen = fdo.OpenSHP(dl, tl);
        }
        /// <summary>
        /// Init as SDF Manager
        /// </summary>
        /// <param name="file"></param>
        /// <param name="read"></param>
        public FDOManager(string file, bool read)
        {
            // backup connection string
            this.sdf_file = file;
            this.sdf_read = read;
            // open connection 
            fdo = new FDOConnector();
            this._isOpen = fdo.OpenSDF(file, read);
        }
        /// <summary>
        /// Initializes Manager with a given FDOConnector
        /// </summary>
        /// <param name="f"></param>
        public FDOManager(FDOConnector f)
        {
            fdo = f;
        }
        ~FDOManager() { try { fdo.Close(); } catch { } }

        /// <summary>
        /// Close the connection
        /// </summary>
        public void Close()
        {
            this._isOpen = false;
            fdo.Close();
        }
        /// <summary>
        /// refresh the connection
        /// </summary>
        public bool ReOpen()
        {
            bool b = true;
            if (!this._isOpen)
            {
                switch (this.source)
                {
                    case FDOProvider.Shape:
                        b = fdo.OpenSHP(this.shp_dl, this.shp_tl);
                        break;
                    case FDOProvider.SDF:
                        b = fdo.OpenSDF(this.sdf_file, this.sdf_read);
                        break;
                }
            }
            return b;
        }
        /// <summary>
        /// Backup the files
        /// </summary>
        /// <param name="className"></param>
        private void Backup(string className)
        {
            if (!this.DoBackup) return;

            int nb;

            if (this.backupedFiles == null)
            {
                // let's create the collection
                this.backupedFiles = new Hashtable();
            }
            if (this.backupedFiles.ContainsKey(className))
            {
                nb = (int)this.backupedFiles[className];    // number of modifications
                nb++;
                this.backupedFiles[className] = nb;
                return;
            }

            if (this.backupDir == null)
            {
                string dirName = fdo.CurrentDirectory + BACKUP_DIR;
                if (!Directory.Exists(dirName))
                {
                    Directory.CreateDirectory(dirName);
                }
                else
                {
                    // let's clean it
                    DirectoryInfo dir = new DirectoryInfo(dirName);
                    DirectoryInfo[] subdirs = dir.GetDirectories();
                    int count = subdirs.Length;
                    foreach (DirectoryInfo d in subdirs)
                    {
                        if (count <= BACKUP_KEEP) 
                            break; 

                        TimeSpan t = DateTime.Now - d.CreationTime;
                        if (t.TotalDays >= BACKUP_DIR_LENGTH)
                        {
                            dir.Delete(true);
                            count--;
                        }
                    }
                }

                dirName += DateTime.Now.Year + "." + DateTime.Now.Month + "." + DateTime.Now.Day + " " +
                            DateTime.Now.Hour + "." + DateTime.Now.Minute + "." + DateTime.Now.Second + "." + DateTime.Now.Millisecond +
                            @"\";
                this.backupDir = Directory.CreateDirectory(dirName);
            }
            foreach (FileInfo f in (new DirectoryInfo(fdo.CurrentDirectory)).GetFiles(className + ".*"))
                f.CopyTo(this.backupDir.FullName + f.Name);

            nb = 0;
            this.backupedFiles.Add(className, nb);
        }
        #endregion

        #region FDO Schema
        /// <summary>
        /// Add a new data property
        /// </summary>
        /// <param name="name">property name</param>
        /// <param name="t">property type</param>
        /// <param name="isNullable">is nullable</param>
        /// <returns>new dataproperty</returns>
        private PropertyDefinition CreateProperty(SHPFilter sf, bool isNullable)
        {
            DataPropertyDefinition d = new DataPropertyDefinition();
            if ( sf.ValueType == typeof(decimal) ||sf.ValueType == typeof(double) )
                d.DataType = DataType.DataType_Decimal;
            else if ( sf.ValueType == typeof(int) || sf.ValueType == typeof(long) )
                d.DataType = DataType.DataType_Int32;       // SHP doesn't support Int64
            else
                d.DataType = DataType.DataType_String;
            d.IsAutoGenerated = false;
            d.IsSystem = false;
            d.Name = sf.Name;
            d.Nullable = isNullable;
            d.ReadOnly = false;
            return d;
        }
        private PropertyDefinition CopyProperty(PropertyDefinition source)
        {
            if (source is GeometricPropertyDefinition)
            {
                return this.CopyGeometricProperty(source as GeometricPropertyDefinition);
            }
            else 
            {
                return this.CopyDataProperty(source as DataPropertyDefinition);
            }
        }
        /// <summary>
        /// Copy DataProperty Attribute by Attribute
        /// </summary>
        /// <param name="sd"></param>
        /// <returns></returns>
        private DataPropertyDefinition CopyDataProperty(DataPropertyDefinition source)
        {
            DataPropertyDefinition p = new DataPropertyDefinition();
            p.AutoDelete = source.AutoDelete;
            p.DataType = source.DataType;
            p.DefaultValue = source.DefaultValue;
            p.Description = source.Description;
            p.IsAutoGenerated = source.IsAutoGenerated;
            p.IsSystem = source.IsSystem;
            p.Length = source.Length;
            p.Name = source.Name;
            p.Nullable = source.Nullable;
            p.Precision = source.Precision;
            p.ReadOnly = source.ReadOnly;
            p.Scale = source.Scale;
            return p;
        }
        /// <summary>
        /// Copy GeometricProperty Attribute by Attribute
        /// </summary>
        /// <param name="sd"></param>
        /// <returns></returns>
        private GeometricPropertyDefinition CopyGeometricProperty(GeometricPropertyDefinition source)
        {
            GeometricPropertyDefinition p = new GeometricPropertyDefinition();
            p.AutoDelete = source.AutoDelete;
            p.Description = source.Description;
            p.GeometryTypes = source.GeometryTypes;
            p.HasElevation = source.HasElevation;
            p.HasMeasure = source.HasMeasure;
            p.IsSystem = source.IsSystem;
            p.Name = source.Name;
            p.ReadOnly = source.ReadOnly;
            return p;
        }
        /// <summary>
        /// Retrieve all classes definitions from default schema
        /// note : the method won't check if the schema exists
        /// </summary>
        private ClassCollection GetDefinition()
        {
            IDescribeSchema descSchema = (IDescribeSchema)fdo.connection.CreateCommand(OSGeo.FDO.Commands.CommandType.CommandType_DescribeSchema);
            FeatureSchemaCollection schemas = descSchema.Execute();
            return schemas[0].Classes;
        }
        /// <summary>
        /// Retrieve class definition
        /// note : the method won't check if the schema & class exist
        /// </summary>
        /// <param name="className">class to analyze</param>
        private ClassDefinition GetDefinition(string className)
        {
            ClassCollection classes = this.GetDefinition();
            return classes.get_Item(classes.IndexOf(className));
        }
        /// <summary>
        /// Check if the given className contains nullable properties aside the ones in "filters"
        /// usefull before inserting data
        /// </summary>
        /// <param name="className"></param>
        /// <param name="filters"></param>
        /// <returns></returns>
        private bool CheckNullable(string className, List<SHPFilter> filters)
        {
            ClassDefinition cd = this.GetDefinition(className);
            foreach (PropertyDefinition p in cd.Properties)
            {
                if (p.Name == "FeatId") break;

                bool inFilter = false;
                foreach (SHPFilter f in filters)
                {
                    if (f.Name == p.Name)
                    {
                        inFilter = true; break;
                    }
                }
                if (!inFilter)
                {
                    if (p.PropertyType == PropertyType.PropertyType_DataProperty)
                    {
                        DataPropertyDefinition d = (DataPropertyDefinition)p;
                        if (!d.Nullable) return false; // this property can't be empty but no data were found for it in "filters"
                    }
                }
            }
            return true;
        }
        /// <summary>
        /// Match two classes definition
        /// </summary>
        /// <param name="class1">Source</param>
        /// <param name="class2">Dest</param>
        /// <param name="filters">Optional columns to check that exist only in dest class</param>
        /// <returns>true if definitions are similar</returns>
        private bool CheckDefinition(string class1, string class2, List<SHPFilter> filters)
        {
            IDescribeSchema descSchema = (IDescribeSchema)fdo.connection.CreateCommand(OSGeo.FDO.Commands.CommandType.CommandType_DescribeSchema);
            FeatureSchemaCollection schemas = descSchema.Execute();
            FeatureSchema schema = schemas[0];
            ClassCollection classes = schema.Classes;
            ClassDefinition cd1 = classes.get_Item(classes.IndexOf(class1));
            ClassDefinition cd2 = classes.get_Item(classes.IndexOf(class2));
            //check if class1 property exists in class2 and if so, ensure compatibility
            foreach (PropertyDefinition p in cd1.Properties)
            {
                foreach (PropertyDefinition p2 in cd2.Properties)
                {
                    // does a property with the same name exists ?
                    if (p.Name == p2.Name)
                    {
                        // let's compare each relevant attribute
                        if (p.PropertyType != p2.PropertyType) return false;
                        switch (p.PropertyType)
                        {
                            case PropertyType.PropertyType_DataProperty:
                                DataPropertyDefinition d1 = (DataPropertyDefinition)p;
                                DataPropertyDefinition d2 = (DataPropertyDefinition)p2;
                                if (d1.DataType != d2.DataType) return false;
                                if (d1.Nullable && !d2.Nullable) return false;
                                // ok, could be more accurate (length / precision and so) but ok for now
                                break;
                            case PropertyType.PropertyType_GeometricProperty:
                                GeometricPropertyDefinition g1 = (GeometricPropertyDefinition)p;
                                GeometricPropertyDefinition g2 = (GeometricPropertyDefinition)p2;
                                if (g1.GeometryTypes != g2.GeometryTypes) return false;
                                // ok, could be more accurate (length / precision and so) but ok for now
                                break;
                        }
                        break;
                    }
                }
            }
            //check if the filters to add exists in class2
            if (filters != null)
            {
                foreach (SHPFilter f in filters)
                {
                    bool fexists = false;
                    foreach (PropertyDefinition p2 in cd2.Properties)
                    {
                        // does a property with the same name exists ?
                        if (f.Name == p2.Name)
                        {
                            fexists = true;
                            break;
                        }
                    }
                    if (!fexists) return false;
                }
            }
            return true;
        }
        /// <summary>
        /// Check for duplicate values
        /// </summary>
        /// <param name="class1">Source</param>
        /// <param name="class2">Dest</param>
        /// <param name="filters">Optional columns to check that exist only in dest class</param>
        /// <param name="keys">Optional Columns that exist in both classes</param>
        /// <returns>true if duplicated data are found</returns>
        private bool CheckDuplicateData(string class1, string class2, List<SHPFilter> filters, List<SHPFilter> keys)
        {
            if( filters == null) filters = new List<SHPFilter>();
            if (keys == null) keys = new List<SHPFilter>();

            if (filters.Count == 0 && keys.Count == 0) return false;

            ClassDefinition definition = GetDefinition(class1);

            bool hasDuplication = false;

            // read source class
            IReader reader = this.Select(class1);

            try
            {                
                while (reader.ReadNext() && !hasDuplication)
                {
                    // read dest class
                    List<SHPFilter> allfilters = new List<SHPFilter>();
                    foreach (SHPFilter sf in keys)
                    {
                        // check if property exist in class1
                        int idx = definition.Properties.IndexOf(sf.Name);
                        if (idx >= 0)
                        {
                            if (!reader.IsNull(sf.Name))
                            {
                                object val = null;
                                DataPropertyDefinition p = (DataPropertyDefinition)definition.Properties[idx];
                                switch (p.DataType)
                                {
                                    case DataType.DataType_Int16: val = reader.GetInt16(p.Name); break;
                                    case DataType.DataType_Int32: val = reader.GetInt32(p.Name); break;
                                    case DataType.DataType_Int64: val = reader.GetInt64(p.Name); break;
                                    case DataType.DataType_Single: val = reader.GetSingle(p.Name); break;
                                    case DataType.DataType_Decimal: val = reader.GetDouble(p.Name); break;
                                    case DataType.DataType_Double: val = reader.GetDouble(p.Name); break;
                                    case DataType.DataType_DateTime: val = reader.GetDateTime(p.Name); break;
                                    case DataType.DataType_Boolean: val = reader.GetBoolean(p.Name); break;
                                    case DataType.DataType_Byte: val = reader.GetByte(p.Name); break;
                                    default:
                                        val = reader.GetString(p.Name); break;
                                }
                                allfilters.Add(new SHPFilter(sf.Name,"=",val));
                                //filter += sf.Name + "=" + (sf.ValueType == typeof(string) ? "'" + val + "'" : val.ToString()) + " AND ";
                            }
                        }
                    }

                    foreach(SHPFilter sf in filters)
                    {
                        // compare with default value mentionned in SHPFilter
                        //filter += sf.Name + "=" + (sf.ValueType == typeof(string) ? "'" + sf.Value + "'" : sf.Value.ToString()) + " AND ";
                        allfilters.Add(new SHPFilter(sf.Name,"=",sf.Value ));
                    }

                    //if(!string.IsNullOrEmpty(filter)) filter = filter.Remove(filter.LastIndexOf(" AND "));
                    
                    IReader reader2 = this.Select(class2,allfilters);

                    if (reader2.ReadNext()) hasDuplication = true;

                    reader2.Close();
                    //reader2.Dispose();
                }
            }
            catch 
            {
                throw;
            }
            finally
            {
                reader.Close();
            }
            return hasDuplication;
        }
        /// <summary>
        /// Retrieve the name of the "Geometry" property
        /// </summary>
        /// <param name="className"></param>
        public string GetGeometryPropertyName(string className)
        {
            string name = string.Empty;
            try
            {
                FeatureClass def = (FeatureClass)this.GetDefinition(className);
                if(def.GeometryProperty != null)
                    name = def.GeometryProperty.Name;
            }
            catch { }
            return name;
        }
        /// <summary>
        /// Check if a specific property can be found in the class
        /// </summary>
        /// <param name="className"></param>
        /// <param name="propertyName"></param>
        /// <returns></returns>
        public bool ContainsPropertyName(string className, string propertyName)
        {
            bool contains = false;
            try
            {
                ClassDefinition def = this.GetDefinition(className);
                foreach (PropertyDefinition p in def.Properties)
                {
                    if (p.Name == propertyName)
                    {
                        contains = true;
                        break;
                    }
                }
            }
            catch { }
            return contains;
        }        

        #endregion
               
        #region FDO Data
        /// <summary>
        /// Copy a class from source to dest
        /// </summary>
        /// <param name="source">Base classname</param>
        /// <param name="dest">Dest classname</param>
        /// <param name="isDataCopy">specify if data must be copied after structure definition</param>
        /// <param name="additionalColumns">list of columns to add (can be empty but not null)</param>
        /// <param name="filter">optional selection filter for data copy</param>
        /// <returns></returns>
        public bool CopyClass(string source, string dest, bool isDataCopy, List<SHPFilter> additionalColumns, List<SHPFilter> filters)
        {
            // get the schema description
            IDescribeSchema descSchema = (IDescribeSchema)fdo.connection.CreateCommand(OSGeo.FDO.Commands.CommandType.CommandType_DescribeSchema);
            FeatureSchemaCollection schemas = descSchema.Execute();
            FeatureSchema schema = schemas[0];

            // copy class def
            ClassCollection classes = schema.Classes;

            // add the dest class
            FeatureClass featdest = new FeatureClass(dest, dest);
            featdest.IsAbstract = false;

            // get source description
            FeatureClass featsource = (FeatureClass)classes.get_Item(classes.IndexOf(source));
            PropertyDefinitionCollection psource = featsource.Properties;
            //DataPropertyDefinitionCollection dsource = featsource.IdentityProperties;

            foreach (PropertyDefinition sp in psource)
            {
                switch (sp.PropertyType)
                {
                    case PropertyType.PropertyType_DataProperty:
                        {
                            DataPropertyDefinition p = this.CopyDataProperty(sp as DataPropertyDefinition);
                            featdest.Properties.Add(p);
                            if (p.IsAutoGenerated)
                                featdest.IdentityProperties.Add(p);
                        }
                        break;
                    case PropertyType.PropertyType_GeometricProperty:
                        {
                            GeometricPropertyDefinition p = this.CopyGeometricProperty(sp as GeometricPropertyDefinition);
                            featdest.Properties.Add(p);
                            featdest.GeometryProperty = p;
                        }
                        break;
                }
            }

            foreach (SHPFilter sf in additionalColumns)
            {
                featdest.Properties.Add( this.CreateProperty(sf,true) );
            }

            // add class
            classes.Add(featdest);

            if (!isDataCopy) return true;            

            // apply schema            
            IApplySchema apply = (IApplySchema)fdo.connection.CreateCommand(OSGeo.FDO.Commands.CommandType.CommandType_ApplySchema);
            apply.FeatureSchema = schema;
            apply.Execute();

            // now select source data
            IReader reader = this.Select(source, filters);

            // insert them in dest class
            IInsert insert = (IInsert)fdo.connection.CreateCommand(OSGeo.FDO.Commands.CommandType.CommandType_Insert);
            insert.SetFeatureClassName(dest);
            try
            {
                while (reader.ReadNext())
                {
                    this.CopyPropertyValues(reader, insert, featsource.Properties);

                    if (additionalColumns.Count > 0)
                    {
                        foreach (SHPFilter sf in additionalColumns)
                            insert.PropertyValues.Add(this.CreatePropertyValue(sf));
                    }
                    insert.Execute();
                    insert.PropertyValues.Clear();
                }
            }
            catch 
            {
                throw;
            }
            finally
            {
                //insert.Dispose();
                reader.Close();
                //reader.Dispose();
            }
            return true;
        }
        /// <summary>
        /// Copy property values from source in reader to iinsert
        /// </summary>
        /// <param name="insert.PropertyValues"></param>
        /// <param name="featsourceProperties"></param>
        private void CopyPropertyValues(IReader reader, IInsert insert, PropertyDefinitionCollection featsourceProperties)
        {
            foreach (PropertyDefinition p in featsourceProperties)
            {
                switch (p.PropertyType)
                {
                    case PropertyType.PropertyType_DataProperty:
                        if (!((DataPropertyDefinition)p).IsAutoGenerated)
                        {
                            if (!reader.IsNull(p.Name))
                            {
                                DataPropertyDefinition ddef = (DataPropertyDefinition)p;
                                switch (ddef.DataType)
                                {
                                    case DataType.DataType_Int16: insert.PropertyValues.Add(new PropertyValue(p.Name, new DataValue(reader.GetInt16(p.Name)))); break;
                                    case DataType.DataType_Int32: insert.PropertyValues.Add(new PropertyValue(p.Name, new DataValue(reader.GetInt32(p.Name)))); break;
                                    case DataType.DataType_Int64: insert.PropertyValues.Add(new PropertyValue(p.Name, new DataValue(reader.GetInt64(p.Name)))); break;
                                    case DataType.DataType_Single: insert.PropertyValues.Add(new PropertyValue(p.Name, new DataValue(reader.GetSingle(p.Name)))); break;
                                    case DataType.DataType_Decimal: insert.PropertyValues.Add(new PropertyValue(p.Name, new DataValue(reader.GetDouble(p.Name), DataType.DataType_Decimal))); break;
                                    case DataType.DataType_Double: insert.PropertyValues.Add(new PropertyValue(p.Name, new DataValue(reader.GetDouble(p.Name), DataType.DataType_Double))); break;
                                    case DataType.DataType_DateTime: insert.PropertyValues.Add(new PropertyValue(p.Name, new DataValue(reader.GetDateTime(p.Name)))); break;
                                    case DataType.DataType_Boolean: insert.PropertyValues.Add(new PropertyValue(p.Name, new DataValue(reader.GetBoolean(p.Name)))); break;
                                    case DataType.DataType_Byte: insert.PropertyValues.Add(new PropertyValue(p.Name, new DataValue(reader.GetByte(p.Name)))); break;
                                    default:
                                        insert.PropertyValues.Add(new PropertyValue(p.Name, new DataValue(reader.GetString(p.Name))));
                                        break; // will probably fail
                                }
                            }
                        }
                        break;
                    case PropertyType.PropertyType_GeometricProperty:
                        insert.PropertyValues.Add(new PropertyValue(p.Name, new GeometryValue(reader.GetGeometry(p.Name))));
                        break;
                }
            }
        }
        /// <summary>
        /// Create a new property value
        /// </summary>
        /// <param name="sf">shp filter</param>
        /// <returns></returns>
        private PropertyValue  CreatePropertyValue(SHPFilter sf)
        {
            if (sf.Name == "Geometry")
            {
                double[] points = (double[])sf.Value;
                // create an FdoGeometryValue for the geometry property value
                GeometryValue geometryValue = new GeometryValue();                
                FgfGeometryFactory geometryFactory = new FgfGeometryFactory();
                // define the external boundary of the polygon
                ILinearRing exteriorRing;
                Int32 nbOrdinates = points.Length;
                exteriorRing = geometryFactory.CreateLinearRing((int)FDODimensionality.Dimensionality_XY, nbOrdinates, points);
                // define the internal ring
                IPolygon poly = geometryFactory.CreatePolygon(exteriorRing, new LinearRingCollection());
                // set the geometry value to this byte array
                Byte[] geometryByteArray = geometryFactory.GetFgf(poly);
                geometryValue.Geometry = geometryByteArray;

                return new PropertyValue(sf.Name, geometryValue);
            }

            if ( sf.ValueType == typeof(decimal) || sf.ValueType == typeof(double) )            
                return new PropertyValue(sf.Name, new DataValue((double)sf.Value, DataType.DataType_Decimal));
             
            if ( sf.ValueType == typeof(int) )
                return new PropertyValue(sf.Name, new DataValue((int)sf.Value)); 

            if( sf.ValueType == typeof(long) )
            {
                // SHP doesn't support Int64
                int i;
                bool b = int.TryParse(sf.Value.ToString(), out i);
                if(b) 
                    return new PropertyValue(sf.Name, new DataValue(i)); 
                else
                    return new PropertyValue(sf.Name, new DataValue( null )); 
            }

            return new PropertyValue(sf.Name, new DataValue(sf.Value.ToString()));
        }
        /// <summary>
        /// Merge two classes data, result is stored in class2
        /// </summary>
        /// <param name="class1"></param>
        /// <param name="class2"></param>
        /// <param name="filters">columns to add</param>
        /// <param name="keys">columns to check for duplicate values</param>
        /// <returns></returns>
        public bool Merge(string class1, string class2, List<SHPFilter> filters, List<SHPFilter> keys)
        {
            try
            {
                if (!CheckDefinition(class1, class2, filters)) return false;
                
                if (keys.Count>0 && CheckDuplicateData(class1, class2, filters, keys)) return false;

                if (!MergeData(class1, class2, filters)) return false;

                return true;
            }
            catch 
            {
                return false;
            }            
        }
        /// <summary>
        /// Merge two classes data (class2 becomes the merged class)
        /// </summary>
        /// <param name="class1"></param>
        /// <param name="class2"></param>
        private bool MergeData(string class1, string class2, List<SHPFilter> filters)
        {
            this.Backup(class2);

            IReader reader = this.Select(class1);
            ClassDefinition definition = GetDefinition(class1);
            IInsert insert = (IInsert)fdo.connection.CreateCommand(OSGeo.FDO.Commands.CommandType.CommandType_Insert);
            insert.SetFeatureClassName(class2);
            try
            {
                while (reader.ReadNext())
                {
                    this.CopyPropertyValues(reader, insert, definition.Properties);

                    // filters
                    if (filters.Count > 0)
                    {
                        foreach (SHPFilter sf in filters)
                            insert.PropertyValues.Add(this.CreatePropertyValue(sf));
                    }
                    insert.Execute();
                    insert.PropertyValues.Clear();
                }
            }
            catch 
            {
                throw;
            }
            finally
            {
                reader.Close();
            }
            return true;
        }
        /// <summary>
        /// Morph a FDO Class to a DataTable object (Geometry not handled)
        /// </summary>
        /// <param name="className"></param>
        /// <returns></returns>
        public DataTable GetDataTable(string className)
        {
            return this.GetDataTable(className, new List<SHPFilter>());
        }
        public DataTable GetDataTable(string className, List<SHPFilter> filters)
        {

            //TODO : cover all data & property types 
            DataTable table = new DataTable(className);

            // step 1 : create the tabledef
            IDescribeSchema descSchema = (IDescribeSchema)fdo.connection.CreateCommand(OSGeo.FDO.Commands.CommandType.CommandType_DescribeSchema);
            FeatureSchemaCollection schemas = descSchema.Execute();
            ClassDefinition classDef = schemas[0].Classes[schemas[0].Classes.IndexOf(className)];
            foreach (PropertyDefinition pdef in classDef.Properties)
            {
                switch (pdef.PropertyType)
                {
                    case PropertyType.PropertyType_GeometricProperty:
                        table.Columns.Add(new DataColumn(pdef.Name, typeof(string)));
                        break;
                    case PropertyType.PropertyType_DataProperty:
                        DataPropertyDefinition ddef = (DataPropertyDefinition)pdef;
                        switch (ddef.DataType)
                        {
                            case DataType.DataType_Int16:
                            case DataType.DataType_Int32:
                            case DataType.DataType_Int64: table.Columns.Add(new DataColumn(pdef.Name, typeof(Int64))); break;
                            case DataType.DataType_Decimal:
                            case DataType.DataType_Double: table.Columns.Add(new DataColumn(pdef.Name, typeof(Double))); break;
                            default:
                                table.Columns.Add(new DataColumn(pdef.Name, typeof(string))); break;
                            //TODO : other types
                        }
                        break;
                }
            }

            // step 2 : populate the table
            IReader reader = this.Select(className,filters);

            try
            {
                while (reader.ReadNext())
                {
                    DataRow row = table.NewRow();
                    foreach (PropertyDefinition pdef in classDef.Properties)
                    {
                        try
                        {
                            switch (pdef.PropertyType)
                            {
                                case PropertyType.PropertyType_GeometricProperty:
                                    row[pdef.Name] = reader.GetGeometry(pdef.Name).Length.ToString();
                                    //int len = reader.GetGeometry(pdef.Name).Length;
                                    //Byte[] geometry = new Byte[len];
                                    //geometry = reader.GetGeometry(pdef.Name);
                                    //tgeo.Nodes.Add(pdef.Name + " : " + len + " octets");
                                    //AnalyzeGeometry(geometry, tgeo);
                                    break;
                                case PropertyType.PropertyType_DataProperty:
                                    if (!reader.IsNull(pdef.Name))
                                    {
                                        DataPropertyDefinition ddef = (DataPropertyDefinition)pdef;
                                        switch (ddef.DataType)
                                        {
                                            case DataType.DataType_Int16: row[pdef.Name] = reader.GetInt16(pdef.Name); break;
                                            case DataType.DataType_Int32: row[pdef.Name] = reader.GetInt32(pdef.Name); break;
                                            case DataType.DataType_Int64: row[pdef.Name] = reader.GetInt64(pdef.Name); break;
                                            case DataType.DataType_Single: row[pdef.Name] = reader.GetSingle(pdef.Name); break;
                                            case DataType.DataType_Decimal:
                                            case DataType.DataType_Double: row[pdef.Name] = reader.GetDouble(pdef.Name); break;
                                            default:
                                                row[pdef.Name] = reader.GetString(pdef.Name); break;
                                        }
                                    }
                                    break;
                            }
                        }
                        catch
                        {
                        }
                    }
                    table.Rows.Add(row);
                }
            }
            catch
            {
                throw;
            }
            finally
            {
                reader.Close();
            }
            return table;
        }

        /// <summary>
        /// Get a class's boundaries (Min(X,Y),Max(X,Y))
        /// </summary>
        /// <param name="className"></param>
        /// <param name="filter"></param>
        /// <returns></returns>
        public Hashtable GetXYEnvelope(string className, List<SHPFilter> filters)
        {
            Hashtable h = new Hashtable();
            h.Add("XMin", double.MaxValue);
            h.Add("YMin", double.MaxValue);
            h.Add("XMax", double.MinValue);
            h.Add("YMax", double.MinValue);
            string geometryName = "Geometry";
            IReader reader = this.Select(className, filters);
            try
            {
                while(!reader.ReadNext())
                {
                    IGeometry geom = this.AnalyzeGeometry(reader.GetGeometry(geometryName));
                    switch (geom.DerivedType)
                    {
                        case GeometryType.GeometryType_Polygon:
                            {
                                if (geom.Envelope.MinX < (double)h["XMin"]) h["XMin"] = geom.Envelope.MinX;
                                if (geom.Envelope.MinY < (double)h["YMin"]) h["YMin"] = geom.Envelope.MinY;
                                if (geom.Envelope.MaxX > (double)h["XMax"]) h["XMax"] = geom.Envelope.MaxX;
                                if (geom.Envelope.MaxY > (double)h["YMax"]) h["YMax"] = geom.Envelope.MaxY;
                            }
                            break;
                    }
                } // next record
            }
            catch{}
            finally
            {
                reader.Close();
                //reader.Dispose();
            }
            return h;
        }
        /// <summary>
        /// Transform byte[] geometry to FDO IGeometry interface
        /// </summary>
        /// <param name="geom_fgf">byte[] = reader.getGeometry(propertyName)</param>
        private IGeometry AnalyzeGeometry(byte[] geom_fgf)
        {
            FgfGeometryFactory gf = new FgfGeometryFactory();
            return gf.CreateGeometryFromFgf(geom_fgf);
        }

        #endregion 

        #region SQL-like methods
        /// <summary>
        /// Select data in the SHP
        /// </summary>
        /// <param name="classname">class</param>
        /// <returns>IReader null or populated</returns>
        public IReader Select(string classname)
        {
            return this.Select(classname, new List<SHPFilter>());
        }
        /// <summary>
        /// Select data in the SHP
        /// </summary>
        /// <param name="classname">class</param>
        /// <param name="filter">filter</param>
        /// <returns>IReader null or populated</returns>
        public IReader Select(string className, List<SHPFilter> filters)
        {
            IReader reader = null;
            try
            {
                ISelect sel = (ISelect)fdo.connection.CreateCommand(OSGeo.FDO.Commands.CommandType.CommandType_Select);
                sel.SetFeatureClassName(className);
                if (filters.Count>0) sel.SetFilter(SHPFilter.ToString(filters));
                reader = sel.Execute();
            }
            catch { }
            return reader;
        }
        /// <summary>
        /// Empty a class
        /// </summary>
        /// <param name="className"></param>
        /// <returns>operation succeeded</returns>
        public bool Delete(string className)
        {            
            return this.Delete(className, new List<SHPFilter>());
        }
        /// <summary>
        /// Delete data in the SHP
        /// </summary>
        /// <param name="className">class</param>
        /// <param name="filter">filter</param>
        /// <returns>operation succeeded</returns>
        public bool Delete(string className, List<SHPFilter> filters)
        {
            this.Backup(className);

            int n = 0;
            try
            {
                IDelete del = (IDelete)fdo.connection.CreateCommand(OSGeo.FDO.Commands.CommandType.CommandType_Delete);
                del.SetFeatureClassName(className);
                if (filters.Count>0) del.SetFilter(SHPFilter.ToString(filters));
                n = del.Execute();
                //del.Dispose();
            } catch {}
            fdo.connection.Flush();
            return (n > 0);
        }
        /// <summary>
        /// Update data in the SHP
        /// </summary>
        /// <param name="className">class</param>
        /// <param name="filter">filter</param>
        /// <param name="updateValues">properties to update</param>
        /// <returns>operation succeeded</returns>
        public bool Update(string className, List<SHPFilter> filters, List<SHPFilter> updateValues)
        {
            this.Backup(className);

            int n = 0;
            try{
                IUpdate update = (IUpdate)fdo.connection.CreateCommand(OSGeo.FDO.Commands.CommandType.CommandType_Update);
                update.SetFeatureClassName(className);
                if (filters.Count>0) update.SetFilter(SHPFilter.ToString(filters));
                PropertyValueCollection pv = update.PropertyValues;
                foreach (SHPFilter f in updateValues)
                {
                    pv.Add(this.CreatePropertyValue(f));
                }
                n = update.Execute();
                //update.Dispose();
            } catch{}
            fdo.connection.Flush();
            return (n>0);
        }
        /// <summary>
        /// Insert
        /// </summary>
        /// <param name="className"></param>
        /// <param name="insertValues"></param>
        /// <param name="checkNullables"></param>
        /// <returns>new FeatId</returns>
        public long Insert(string className, List<SHPFilter> insertValues)
        {
            // if(!this.CheckNullable(className, insertValues)) return 0; // some not nullable property won't be assigned

            this.Backup(className);

            long featid = 0;
            try{
                IInsert insert = (IInsert)fdo.connection.CreateCommand(OSGeo.FDO.Commands.CommandType.CommandType_Insert);
                insert.SetFeatureClassName(className);
                PropertyValueCollection pv = insert.PropertyValues;
                foreach (SHPFilter f in insertValues)
                {
                    pv.Add(this.CreatePropertyValue(f));
                }
                IReader reader = insert.Execute();                
                //insert.Dispose();
                if (reader != null)
                {
                    if (reader.ReadNext())
                    {
                        featid = reader.GetInt32("FeatId");
                    }
                    reader.Close();
                    //reader.Dispose();
                }
            }
            catch { }
            fdo.connection.Flush();
            return featid;
        }
        #endregion 



    }

    // TODO : handle multiple value types
    public class SHPFilter
    {
        public string Name;
        public string Operator;
        public object Value;
        public Type ValueType;
        /// <summary>
        /// SHPFilter
        /// </summary>
        /// <param name="p">property name</param>
        /// <param name="op">operand</param>
        /// <param name="v">value</param>
        public SHPFilter(string p, string op, object v)
        {
            this.Name = p;
            this.Operator = op;
            this.Value = v;
            this.ValueType = v.GetType();
        }
        /// <summary>
        /// SHPFilter
        /// </summary>
        /// <param name="p">property name</param>
        /// <param name="v">value</param>
        public SHPFilter(string p, object v)
        {
            this.Name = p;
            this.Operator = string.Empty;
            this.Value = v;
            this.ValueType = v.GetType();
        }
        public override string ToString()
        {
            return this.Name + this.Operator + (this.ValueType == typeof(string) ? "'" + this.Value.ToString() + "'" : this.Value);
        }

        public static string ToString(List<SHPFilter> filters)
        {
            string s = string.Empty;
            foreach (SHPFilter f in filters)
            {
                s += (!string.IsNullOrEmpty(s) ? " AND " : "") + f.ToString();
            }
            return s;
        }
    }

}

